from typing import Optional

from PySide2.QtCore import Qt
from PySide2.QtWidgets import QWidget, QVBoxLayout, QLineEdit, QTreeWidget, QTreeWidgetItem, QPushButton

from angr.analyses.decompiler.decompilation_options import options as dec_options
from angr.analyses.decompiler.optimization_passes import get_optimization_passes, get_default_optimization_passes
from angr.analyses.decompiler.peephole_optimizations import EXPR_OPTS, STMT_OPTS


class OptionType:
    OPTION = 1
    OPTIMIZATION_PASS = 2
    PEEPHOLE_OPTIMIZATION = 3


class QDecompilationOption(QTreeWidgetItem):
    def __init__(self, parent, option, type_: int, enabled=True):
        super().__init__(parent)
        self.option = option
        self.type = type_

        if self.type == OptionType.OPTIMIZATION_PASS:
            self.setText(0, option.NAME)
            self.setToolTip(0, option.DESCRIPTION)
        elif self.type == OptionType.OPTION:
            self.setText(0, option.NAME)
            self.setToolTip(0, option.DESCRIPTION)
        elif self.type == OptionType.PEEPHOLE_OPTIMIZATION:
            self.setText(0, option.NAME)
            self.setToolTip(0, option.DESCRIPTION)
        else:
            raise NotImplementedError("Unsupported option type %s." % self.type_)

        self.setFlags(self.flags() | Qt.ItemIsUserCheckable)
        if enabled:
            self.setCheckState(0, Qt.Checked)
        else:
            self.setCheckState(0, Qt.Unchecked)


class QDecompilationOptions(QWidget):
    def __init__(self, code_view, instance, options=None, passes=None, peephole_opts=None):
        super().__init__()

        self.dirty = True

        self._code_view = code_view
        self._instance = instance
        self._options = options
        self._opti_passes = passes
        self._peephole_opts = peephole_opts

        # widgets
        self._search_box = None  # type:QLineEdit
        self._treewidget = None  # type:QTreeWidget
        self._apply_btn = None  # type:QPushButton

        self._qoptions = [ ]
        self._qoptipasses = [ ]
        self._qpeephole_opts = [ ]

        self._init_widgets()

        self.reload()

    def reload(self, force=False):
        if force or self._options is None:
            self._options = self.get_default_options()

        if force or self._opti_passes is None:
            if not self._instance.project.am_none:
                self._opti_passes = self.get_all_passes()
            else:
                self._opti_passes = []

        if force or self._peephole_opts is None:
            self._peephole_opts = self.get_all_peephole_opts()

        self._reload_options()

    def _on_item_changed(self, item, _column):
        if getattr(item.option, 'clears_cache', True):
            self.dirty = True

    def _on_apply_pressed(self):
        if self.dirty:
            self.dirty = False
            # clear the cached version
            self._code_view.decompile(reset_cache=True)
        else:
            self._code_view.codegen.reapply_options(self.option_and_values)
            self._code_view.codegen.am_event()

    @property
    def selected_passes(self):
        selected = [ ]
        for item in self._qoptipasses:
            if item.checkState(0):
                selected.append(item.option)
        return selected

    @property
    def selected_peephole_opts(self):
        selected = []
        for item in self._qpeephole_opts:
            if item.checkState(0):
                selected.append(item.option)
        return selected

    @property
    def option_and_values(self):
        ov = [ ]
        for item in self._qoptions:
            if item.checkState(0):
                ov.append((item.option, True))
            else:
                ov.append((item.option, False))
        return ov

    @property
    def options(self):
        return self._options

    @options.setter
    def options(self, v):
        self._options = v
        self._reload_options()

    def get_default_options(self):  # pylint: disable=no-self-use
        return dec_options

    def get_default_passes(self):
        if self._instance is None or self._instance.project.am_none:
            return set()
        return get_default_optimization_passes(self._instance.project.arch, self._instance.project.simos.name)

    def get_all_passes(self):
        return get_optimization_passes(self._instance.project.arch, self._instance.project.simos.name)

    def get_default_peephole_opts(self):  # pylint: disable=no-self-use
        return STMT_OPTS + EXPR_OPTS

    def get_all_peephole_opts(self):  # pylint: disable=no-self-use
        return STMT_OPTS + EXPR_OPTS

    def _init_widgets(self):

        # search box
        self._search_box = QLineEdit()
        self._search_box.textChanged.connect(self._on_search_box_text_changed)

        # tree view
        self._treewidget = QTreeWidget()
        self._treewidget.setHeaderHidden(True)
        self._treewidget.itemChanged.connect(self._on_item_changed)

        # refresh button
        self._apply_btn = QPushButton("Apply")
        self._apply_btn.clicked.connect(self._on_apply_pressed)

        layout = QVBoxLayout()
        layout.addWidget(self._search_box)
        layout.addWidget(self._treewidget)
        layout.addWidget(self._apply_btn)

        self.setLayout(layout)

    def _reload_options(self, filter_by: Optional[str]=None):

        self._treewidget.clear()
        self._qoptions.clear()
        self._qoptipasses.clear()
        self._qpeephole_opts.clear()

        categories = { }

        if filter_by:
            filter_by = filter_by.lower()

        # populate the tree widget with new options
        for option in sorted(self._options, key=lambda x: x.NAME):
            if filter_by and not (filter_by in option.NAME.lower() or filter_by in option.category.lower()):
                continue
            if option.category in categories:
                category = categories[option.category]
            else:
                category = QTreeWidgetItem(self._treewidget, [option.category])
                categories[option.category] = category

            w = QDecompilationOption(category, option, OptionType.OPTION, enabled=option.default_value)
            self._qoptions.append(w)

        passes_category = QTreeWidgetItem(self._treewidget, ["Optimization Passes"])
        categories['passes'] = passes_category

        default_passes = set(self.get_default_passes())
        for pass_ in self._opti_passes:
            if filter_by and not (filter_by in pass_.__name__.lower() or filter_by in pass_.NAME.lower()):
                continue
            w = QDecompilationOption(passes_category, pass_, OptionType.OPTIMIZATION_PASS,
                                     enabled=pass_ in default_passes)
            self._qoptipasses.append(w)

        po_category = QTreeWidgetItem(self._treewidget, ["Peephole Optimizations"])
        categories['peephole_opts'] = po_category

        default_peephole_opts = self.get_default_peephole_opts()
        for opt_ in self._peephole_opts:
            if filter_by and not (filter_by in opt_.NAME.lower() or filter_by in opt_.DESCRIPTION.lower()):
                continue
            w = QDecompilationOption(po_category, opt_, OptionType.PEEPHOLE_OPTIMIZATION,
                                     enabled=opt_ in default_peephole_opts)
            self._qpeephole_opts.append(w)

        # expand all
        self._treewidget.expandAll()

    def _on_search_box_text_changed(self, text: str):
        self._reload_options(filter_by=text)



from typing import Tuple, Dict, Optional
import os
import binascii
import logging
import archinfo
from cle import Blob

from PySide2.QtWidgets import QDialog, QVBoxLayout, QHBoxLayout, QLabel, QTabWidget, QCheckBox, QFrame, QGroupBox, \
    QListWidgetItem, QListWidget, QMessageBox, QLineEdit, QGridLayout, QComboBox, QSizePolicy, QDialogButtonBox
from PySide2.QtCore import Qt


l = logging.getLogger('dialogs.load_binary')


class LoadBinaryError(Exception):
    """
    An error loading the binary.
    """


class LoadBinary(QDialog):
    """
    Dialog displaying loading options for a binary.
    """

    def __init__(self, partial_ld, parent=None):
        super().__init__(parent)

        # initialization
        self.file_path = partial_ld.main_object.binary
        self.md5 = None
        self.sha256 = None
        self.option_widgets = { }
        self.is_blob = isinstance(partial_ld.main_object, Blob)
        self.arch = partial_ld.main_object.arch

        # return values
        self.cfg_args = None
        self.load_options = None
        self.variable_recovery_args = None

        self.setWindowTitle('Load a new binary')

        # checksums
        if hasattr(partial_ld.main_object, 'md5') and partial_ld.main_object.md5 is not None:
            self.md5 = binascii.hexlify(partial_ld.main_object.md5).decode("ascii")
        if hasattr(partial_ld.main_object, 'sha256') and partial_ld.main_object.sha256 is not None:
            self.sha256 = binascii.hexlify(partial_ld.main_object.sha256).decode("ascii")

        self.main_layout = QVBoxLayout()

        self._init_widgets()

        self._try_loading(partial_ld)

        self.setLayout(self.main_layout)

    @property
    def filename(self):
        return os.path.basename(self.file_path)

    #
    # Private methods
    #

    def _try_loading(self, partial_ld):
        deps = [ ]
        processed_objects = set()
        for ident, obj in partial_ld._satisfied_deps.items():
            if obj is partial_ld._kernel_object or \
                    obj is partial_ld._extern_object or \
                    obj is partial_ld.main_object:
                continue
            if obj in processed_objects:
                continue
            deps.append(ident)
            processed_objects.add(obj)

        # dependencies

        dep_list = self.option_widgets['dep_list']  # type: QListWidget
        for dep in deps:
            dep_item = QListWidgetItem(dep)
            dep_item.setData(Qt.CheckStateRole, Qt.Unchecked)
            dep_list.addItem(dep_item)

    def _init_widgets(self):

        layout = QGridLayout()
        self.main_layout.addLayout(layout)

        # filename

        filename_caption = QLabel(self)
        filename_caption.setText('File name:')

        filename = QLabel(self)
        filename.setText(self.filename)

        layout.addWidget(filename_caption, 0, 0, Qt.AlignRight)
        layout.addWidget(filename, 0, 1)

        # md5

        if self.md5 is not None:
            md5_caption = QLabel(self)
            md5_caption.setText('MD5:')
            md5 = QLineEdit(self)
            md5.setText(self.md5)
            md5.setReadOnly(True)

            layout.addWidget(md5_caption, 1, 0, Qt.AlignRight)
            layout.addWidget(md5, 1, 1)

        # sha256

        if self.sha256 is not None:
            sha256_caption = QLabel(self)
            sha256_caption.setText('SHA256:')
            sha256 = QLineEdit(self)
            sha256.setText(self.sha256)
            sha256.setReadOnly(True)

            layout.addWidget(sha256_caption, 2, 0, Qt.AlignRight)
            layout.addWidget(sha256, 2, 1)

        # central tab

        tab = QTabWidget()
        self._init_central_tab(tab)

        self.main_layout.addWidget(tab)

        # buttons

        buttons = QDialogButtonBox(parent=self)
        buttons.setStandardButtons(QDialogButtonBox.StandardButton.Cancel | QDialogButtonBox.StandardButton.Ok)
        buttons.accepted.connect(self._on_ok_clicked)
        buttons.rejected.connect(self._on_cancel_clicked)
        self.main_layout.addWidget(buttons)

    def _init_central_tab(self, tab):
        self._init_load_options_tab(tab)
        self._init_analysis_options_tab(tab)

    def _init_load_options_tab(self, tab):
        arch_layout = QHBoxLayout()
        arch_caption = QLabel(self)
        arch_caption.setText('Architecture:')
        arch_caption.setSizePolicy(QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed))
        arch_layout.addWidget(arch_caption)
        arch_combo = QComboBox(self)
        for arch in archinfo.all_arches:
            addendum = ' (P-code Engine)' if hasattr(arch, 'pcode_arch') else ''
            arch_combo.addItem(f'{arch.bits}b {arch.name} ({arch.memory_endness[-2:]}){addendum}', str(arch))
        index = arch_combo.findData(str(self.arch))
        arch_combo.setCurrentIndex(index)
        arch_layout.addWidget(arch_combo)
        self.option_widgets['arch'] = arch_combo

        if self.is_blob:
            blob_layout = QGridLayout()

            # load address
            base_addr_caption = QLabel(self)
            base_addr_caption.setText('Base Address:')
            blob_layout.addWidget(base_addr_caption, 1, 0)
            base_addr = QLineEdit(self)
            base_addr.setText('0')
            blob_layout.addWidget(base_addr, 1, 1)
            self.option_widgets['base_addr'] = base_addr

            # entry address
            entry_addr_caption = QLabel(self)
            entry_addr_caption.setText('Entry Address:')
            blob_layout.addWidget(entry_addr_caption, 2, 0)
            entry_addr = QLineEdit(self)
            entry_addr.setText('0')
            blob_layout.addWidget(entry_addr, 2, 1)
            self.option_widgets['entry_addr'] = entry_addr

        # load debug symbols
        load_debug_info = QCheckBox()
        load_debug_info.setText("Load debug information if available")
        load_debug_info.setChecked(True)
        self.option_widgets['load_debug_info'] = load_debug_info

        # auto load libs

        auto_load_libs = QCheckBox()
        auto_load_libs.setText("Automatically load all libraries (slow, not recommended)")
        auto_load_libs.setChecked(False)
        self.option_widgets['auto_load_libs'] = auto_load_libs

        # dependencies list

        dep_group = QGroupBox("Dependencies")
        dep_list = QListWidget()
        self.option_widgets['dep_list'] = dep_list

        sublayout = QVBoxLayout()
        sublayout.addWidget(dep_list)
        dep_group.setLayout(sublayout)

        layout = QVBoxLayout()
        if self.is_blob:
            layout.addLayout(blob_layout)
        layout.addLayout(arch_layout)
        layout.addWidget(load_debug_info)
        layout.addWidget(auto_load_libs)
        layout.addWidget(dep_group)
        layout.addStretch(0)

        frame = QFrame(self)
        frame.setLayout(layout)
        tab.addTab(frame, "Loading Options")

    def _init_analysis_options_tab(self, tab):
        resolve_indirect_jumps = QCheckBox(self)
        resolve_indirect_jumps.setText('Resolve indirect jumps')
        resolve_indirect_jumps.setChecked(True)
        self.option_widgets['resolve_indirect_jumps'] = resolve_indirect_jumps

        collect_data_refs = QCheckBox(self)
        collect_data_refs.setText('Collect cross references and guess data types')
        collect_data_refs.setChecked(True)
        self.option_widgets['data_references'] = collect_data_refs

        xrefs = QCheckBox(self)
        xrefs.setText('Deep analysis on cross references (slow)')
        xrefs.setChecked(False)
        self.option_widgets['cross_references'] = xrefs

        recover_variables_on_matched_functions = QCheckBox(self)
        recover_variables_on_matched_functions.setText("Recover variables on signature-matched functions")
        recover_variables_on_matched_functions.setChecked(False)
        self.option_widgets['recover_variables_on_matched_functions'] = recover_variables_on_matched_functions

        skip_unmapped_addrs = QCheckBox()
        skip_unmapped_addrs.setText("Skip unmapped addresses")
        skip_unmapped_addrs.setChecked(True)
        self.option_widgets['skip_unmapped_addrs'] = skip_unmapped_addrs

        layout = QVBoxLayout()
        layout.addWidget(resolve_indirect_jumps)
        layout.addWidget(collect_data_refs)
        layout.addWidget(xrefs)
        layout.addWidget(recover_variables_on_matched_functions)
        layout.addWidget(skip_unmapped_addrs)
        layout.addStretch(0)
        frame = QFrame(self)
        frame.setLayout(layout)
        tab.addTab(frame, 'Analysis Options')

    #
    # Event handlers
    #

    def _on_ok_clicked(self):

        force_load_libs = [ ]
        skip_libs = set()

        dep_list = self.option_widgets['dep_list']  # type: QListWidget
        for i in range(dep_list.count()):
            item = dep_list.item(i)  # type: QListWidgetItem
            if item.checkState() == Qt.Checked:
                force_load_libs.append(item.text())
            else:
                skip_libs.add(item.text())

        self.load_options = { }
        self.load_options['auto_load_libs'] = self.option_widgets['auto_load_libs'].isChecked()
        self.load_options['load_debug_info'] = self.option_widgets['load_debug_info'].isChecked()
        self.load_options['arch'] = archinfo.all_arches[self.option_widgets['arch'].currentIndex()]

        if self.is_blob:
            self.load_options['main_opts'] = {
                'backend': 'blob',
                'base_addr': int(self.option_widgets['base_addr'].text(), 16),
                'entry_point': int(self.option_widgets['entry_addr'].text(), 16),
            }

        if force_load_libs:
            self.load_options['force_load_libs'] = force_load_libs
        if skip_libs:
            self.load_options['skip_libs'] = skip_libs

        self.cfg_args = {
            'resolve_indirect_jumps': self.option_widgets['resolve_indirect_jumps'].isChecked(),
            'data_references': self.option_widgets['data_references'].isChecked(),
            'cross_references': self.option_widgets['cross_references'].isChecked(),
            'skip_unmapped_addrs': self.option_widgets['skip_unmapped_addrs'].isChecked()
        }

        self.variable_recovery_args = {
            'skip_signature_matched_functions':
                not self.option_widgets['recover_variables_on_matched_functions'].isChecked(),
        }

        self.close()

    def _on_cancel_clicked(self):
        self.cfg_args = None
        self.close()

    @staticmethod
    def run(partial_ld) -> Tuple[Optional[Dict],Optional[Dict],Optional[Dict]]:
        try:
            dialog = LoadBinary(partial_ld)
            dialog.setModal(True)
            dialog.exec_()

            if dialog.cfg_args is not None:
                # load the binary
                return dialog.load_options, dialog.cfg_args, dialog.variable_recovery_args
        except LoadBinaryError:
            pass
        return None, None, None

    @staticmethod
    def binary_arch_detect_failed(filename:str, archinfo_msg:str):
        # TODO: Normalize the path for Windows
        QMessageBox.warning(None,
                            "Architecture selection failed",
                            f"{archinfo_msg} for binary:\n\n{filename}")

    @staticmethod
    def binary_loading_failed(filename):
        # TODO: Normalize the path for Windows
        QMessageBox.critical(None,
                             "Failed to load binary",
                             f"angr failed to load binary {filename}.")



__version__ = "9.2.1"

# Hack used to work around the slow-responsiveness issue with the GUI
# PySide2 5.14.2 solves this problem but it introduces other bugs
# See https://bugreports.qt.io/browse/PYSIDE-803
try:
    import PySide2
    version = [int(k) for k in PySide2.__version__.split(".")[:4]]
    while len(version) < 4:
        version.append(0)
    version = tuple(version)
    if version < (5, 14, 2, 0):
        import sys
        sys.setswitchinterval(0.00001)
except ImportError:
    pass


try:
    # make sure qtpy (which is used in PyQodeNG.core) is using PySide2
    import os
    os.environ['QT_API'] = 'pyside2'
    import qtpy
except ImportError:
    # qtpy is not installed
    pass



import logging
from typing import Type, List

from PySide2.QtWidgets import QDialog, QVBoxLayout, QPushButton, QFrame, QGroupBox, QListWidgetItem, \
    QListWidget, QFileDialog, QMessageBox, QDialogButtonBox
from PySide2.QtCore import Qt

from angrmanagement.plugins import load_plugins_from_file

_l = logging.getLogger(__name__)


class QPluginListWidgetItem(QListWidgetItem):
    """
    Plugin list item.
    """

    def __init__(self, plugin_cls, **kwargs):
        super().__init__(**kwargs)
        self.plugin_class = plugin_cls  # type: Type[BasePlugin]
        self.setText(plugin_cls.get_display_name())


# TODO: Add plugin settings, reloading, etc.

class LoadPlugins(QDialog):
    """
    Dialog to display loaded plugins, enable/disable plugins, and load new plugins.
    """

    def __init__(self, plugin_mgr, parent=None):
        super().__init__(parent)

        self._pm = plugin_mgr  # type: PluginManager
        self._installed_plugin_list = None  # type: QListWidget

        self.setWindowTitle('Installed Plugins')
        self.main_layout = QVBoxLayout()

        self._init_widgets()
        self._populate_installed_plugin_list()
        self.setLayout(self.main_layout)

    #
    # Private methods
    #

    def _init_plugin_list(self):
        plugin_group = QGroupBox("Plugins")
        self._installed_plugin_list = QListWidget(self)

        sublayout = QVBoxLayout()
        sublayout.addWidget(self._installed_plugin_list)
        plugin_group.setLayout(sublayout)

        layout = QVBoxLayout()
        # layout.addWidget(auto_load_libs)
        layout.addWidget(plugin_group)
        layout.addStretch(0)

        frame = QFrame(self)
        frame.setLayout(layout)

        self.main_layout.addWidget(frame)

    def _populate_installed_plugin_list(self):
        for cls in self._pm.loaded_plugins:
            plugin_item = QPluginListWidgetItem(plugin_cls=cls)
            if self._pm.get_plugin_instance(cls) is not None:
                plugin_item.setCheckState(Qt.Checked)
            else:
                plugin_item.setCheckState(Qt.Unchecked)
            self._installed_plugin_list.addItem(plugin_item)

    def _init_widgets(self):
        load_button = QPushButton(self)
        load_button.setText("Load Plugin")
        load_button.clicked.connect(self._on_load_clicked)
        self.main_layout.addWidget(load_button)

        self._init_plugin_list()

        buttons = QDialogButtonBox(parent=self)
        buttons.setStandardButtons(QDialogButtonBox.StandardButton.Cancel | QDialogButtonBox.StandardButton.Ok)
        buttons.accepted.connect(self._on_ok_clicked)
        buttons.rejected.connect(self.close)
        self.main_layout.addWidget(buttons)

    #
    # Event handlers
    #

    def _on_ok_clicked(self):
        list_items = self._installed_plugin_list.findItems('*', Qt.MatchWildcard)  # type: List[QPluginListWidgetItem]
        for i in list_items:
            checked = i.checkState() == Qt.Checked

            if checked and self._pm.get_plugin_instance(i.plugin_class) is None:
                self._pm.activate_plugin(i.plugin_class)
            elif not checked and self._pm.get_plugin_instance(i.plugin_class) is not None:
                self._pm.deactivate_plugin(i.plugin_class)

        self._pm.save_enabled_plugins_to_config()
        self.close()

    def _on_load_clicked(self):
        file_path, _ = QFileDialog.getOpenFileName(
            self, "Open a plugin (select __init__.py for packages)", "", "Python files (*.py)")
        if not file_path:
            return
        plugins = load_plugins_from_file(file_path)

        if not plugins:
            QMessageBox.warning(self, "Error", "File contained no plugins")
            return

        errors = [x for x in plugins if isinstance(x, Exception)]
        if errors:
            QMessageBox.warning(self, "Error", "Loading errored with %s" % errors[0])
            return

        for plugin in plugins:
            plugin_item = QPluginListWidgetItem(plugin_cls=plugin)
            plugin_item.setCheckState(Qt.Unchecked)
            self._pm.load_plugin(plugin)
            self._installed_plugin_list.addItem(plugin_item)



import functools
from typing import Optional, List, Type, Union, TYPE_CHECKING
import logging
import os

from PySide2.QtGui import QColor

from ..config import config_path
from ..config.config_manager import ENTRIES
from ..ui.menus.menu import MenuEntry, MenuSeparator
from ..ui.toolbars.toolbar import ToolbarAction
from ..daemon.url_handler import register_url_action, UrlActionBinaryAware
from ..daemon.client import DaemonClient
from ..ui.widgets.qblock import QBlock
from ..config import Conf, save_config
from . import load_plugins_from_dir
from .base_plugin import BasePlugin

if TYPE_CHECKING:
    from ..ui.workspace import Workspace


l = logging.getLogger(__name__)

# The plugin manager can be initialized in two modes:
# - UI mode, where workspace is not None
# - headless mode, where workspace is None, and any plugin that requires workspace will not be initialized
#
# Plugins can be in two states:
# - Loaded, but inactive. In this, the class is present in .loaded_plugins but there is nothing in .active_plugins
# - Activated. In this, there is an instance of the class in .active_plugins
#
# ...so this class has functions to transition plugins into and between these states.
#
# The actual process of excavating a process from the filesystem into a python class ("loading") needs to be dealt
# with before anything touches this class. There are functions to do that in the plugins package but they need to be
# tied to the user's settings related to loading paths and activation. Presently this is split between MainWindow (the
# first-boot autoload part) and the LoadPlugins dialog (the extra loading and tweaking activation)

class PluginManager:
    def __init__(self, workspace: Optional['Workspace']):
        self.workspace = workspace
        # should one or both of these be ObjectContainers? I think no since we should be synchronizing on models, not
        # views/controllers. not super clear... that's not a hard and fast rule
        self.loaded_plugins = []  # type: List[Type[BasePlugin]]
        self.active_plugins = []  # type: List[BasePlugin]

    def discover_and_initialize_plugins(self):
        os.environ['AM_BUILTIN_PLUGINS'] = os.path.dirname(__file__)
        enabled_plugins = [ plugin_.strip() for plugin_ in Conf.enabled_plugins.split(',') if plugin_.strip() ]
        for search_dir in Conf.plugin_search_path.split(':'):
            search_dir = os.path.expanduser(search_dir)
            search_dir = os.path.expandvars(search_dir)
            for plugin_or_exception in load_plugins_from_dir(search_dir):
                if isinstance(plugin_or_exception, Exception):
                    l.warning("Exception occurred during plugin loading: %s", plugin_or_exception)
                else:
                    plugin_or_exception: Type[BasePlugin]
                    plugin_conf_key = "plugin_%s_enabled" % plugin_or_exception.__name__

                    # see if we can't load this plugin because headless mode
                    if self.workspace is None and plugin_or_exception.REQUIRE_WORKSPACE:
                        # still note that we can use the url handlers
                        for action in plugin_or_exception.URL_ACTIONS:
                            register_url_action(action, UrlActionBinaryAware)
                    # see if the plugin is enabled or not
                    elif any(plugin in repr(plugin_or_exception) for plugin in enabled_plugins) and \
                            not (hasattr(Conf, plugin_conf_key) and getattr(Conf, plugin_conf_key) is False):
                        self.activate_plugin(plugin_or_exception)
                    else:
                        plugin_or_exception: Type[BasePlugin]
                        if (plugin_or_exception.REQUIRE_WORKSPACE and self.workspace is not None) \
                                or not plugin_or_exception.REQUIRE_WORKSPACE:
                            self.load_plugin(plugin_or_exception)
                            l.info("Blacklisted plugin %s", plugin_or_exception.get_display_name())

    def load_plugin(self, plugin_cls: Type[BasePlugin]):
        if plugin_cls in self.loaded_plugins:
            return
        if type(plugin_cls) is not type or not issubclass(plugin_cls, BasePlugin):
            raise TypeError("Cannot load a plugin which is not a BasePlugin subclass")
        if hasattr(plugin_cls, '_%s__i_hold_this_abstraction_token' % plugin_cls.__name__):
            raise TypeError("Cannot load an abstract plugin")
        if plugin_cls.REQUIRE_WORKSPACE and self.workspace is None:
            raise RuntimeError("Cannot load plugin %s in headless mode.")
        self.loaded_plugins.append(plugin_cls)

    def activate_plugin(self, plugin_cls: Type[BasePlugin]):
        self.load_plugin(plugin_cls)  # just to be sure, also perform the sanity checks
        if self.get_plugin_instance(plugin_cls) is not None:
            return

        try:
            plugin = plugin_cls(self.workspace)
            self.active_plugins.append(plugin)
            plugin.__cached_status_bar_widgets = []
            plugin.__cached_toolbar_actions = []  # a hack, lol. really this could be a mapping on PluginManager but idc
            plugin.__cached_menu_actions = []  # as above

            if self.workspace is not None:
                self._register_status_bar_widgets(plugin)
                self._register_toolbar_actions(plugin_cls, plugin)
                self._register_menu_buttons(plugin_cls, plugin)

                for dview in self.workspace.view_manager.views_by_category['disassembly']:
                    plugin.instrument_disassembly_view(dview)
                for cview in self.workspace.view_manager.views_by_category['pseudocode']:
                    plugin.instrument_code_view(cview)

                self._register_url_actions(plugin_cls, plugin)
                self._register_configuration_entries(plugin_cls)

            for action in plugin_cls.URL_ACTIONS:
                register_url_action(action, UrlActionBinaryAware)

        except Exception: #pylint: disable=broad-except
            l.warning("Plugin %s failed to activate:", plugin_cls.get_display_name(),
                      exc_info=True)
        else:
            l.info("Activated plugin %s", plugin_cls.get_display_name())

    def save_enabled_plugins_to_config(self):
        # pylint: disable=assigning-non-slot
        Conf.enabled_plugins = ','.join(p.__class__.__name__ for p in self.active_plugins)
        save_config()

    def _register_status_bar_widgets(self, plugin: BasePlugin) -> None:
        gen = plugin.status_bar_permanent_widgets()
        if gen is not None:
            for widget in gen:
                self.workspace.main_window.statusBar().addPermanentWidget(widget)
                widget.show()
            plugin.__cached_status_bar_widgets.append(widget)

    def _register_toolbar_actions(self, plugin_cls: Type[BasePlugin], plugin: BasePlugin) -> None:
        for idx, (icon, tooltip) in enumerate(plugin_cls.TOOLBAR_BUTTONS):
            action = ToolbarAction(icon, 'plugin %s toolbar %d' % (plugin_cls, idx), tooltip,
                                   functools.partial(self._dispatch_single, plugin, BasePlugin.handle_click_toolbar,
                                                     False, idx))
            plugin.__cached_toolbar_actions.append(action)
            self.workspace._main_window._file_toolbar.add(action)

    def _register_menu_buttons(self, plugin_cls: Type[BasePlugin], plugin: BasePlugin) -> None:
        if plugin_cls.MENU_BUTTONS:
            self.workspace._main_window._plugin_menu.add(MenuSeparator())
        for idx, text in enumerate(plugin_cls.MENU_BUTTONS):
            action = MenuEntry(text,
                               functools.partial(self._dispatch_single, plugin, BasePlugin.handle_click_menu, False,
                                                 idx))
            plugin.__cached_menu_actions.append(action)
            self.workspace._main_window._plugin_menu.add(action)

    def _register_url_actions(self, plugin_cls: Type[BasePlugin], plugin: BasePlugin) -> None:
        for action in plugin_cls.URL_ACTIONS:
            DaemonClient.register_handler(action,
                                          functools.partial(self._dispatch_single,
                                                            plugin,
                                                            BasePlugin.handle_url_action,
                                                            False,
                                                            action
                                                            )
                                          )

    def _register_configuration_entries(self, plugin_cls: Type[BasePlugin]) -> None:
        new_entries_added = False
        for ent in plugin_cls.CONFIG_ENTRIES:
            if ent not in ENTRIES:
                # if a plugin is disabled and then enabled, the entry may already exist?
                ENTRIES.append(ent)
                new_entries_added = True

        if new_entries_added:
            # reload configuration manager so that it's aware of newly added entries
            Conf.load_initial_entries(reset=False)
            # reload the configuration file
            new_conf = Conf.parse_file(config_path)
            Conf._entries = new_conf._entries

    def get_plugin_instance_by_name(self, plugin_cls_name: str) -> Optional[BasePlugin]:
        instances = [plugin for plugin in self.active_plugins if plugin.__class__.__name__.split(".")[-1] == plugin_cls_name]
        if not instances:
            return None
        if len(instances) > 1:
            l.error("Somehow there is more than one instance of %s active?", plugin_cls_name)
        return instances[0]

    def get_plugin_instance(self, plugin_cls: Type[BasePlugin]) -> Optional[BasePlugin]:
        instances = [plugin for plugin in self.active_plugins if type(plugin) is plugin_cls]
        if len(instances) == 0:
            return None
        if len(instances) > 1:
            l.error("Somehow there is more than one instance of %s active?", plugin_cls.get_display_name())
        return instances[0]

    def deactivate_plugin(self, plugin: Union[BasePlugin, Type[BasePlugin]]):
        # this method should work on both instances and classes
        if type(plugin) is type:
            plugin = self.get_plugin_instance(plugin)

        if plugin not in self.active_plugins:
            return

        for widget in plugin.__cached_status_bar_widgets:
            self.workspace.main_window.statusBar().removeWidget(widget)
        for action in plugin.__cached_toolbar_actions:
            self.workspace._main_window._file_toolbar.remove(action)
        for action in plugin.__cached_menu_actions:
            self.workspace._main_window._plugin_menu.remove(action)

        try:
            plugin.teardown()

        except Exception: #pylint: disable=broad-except
            l.warning("Plugin %s errored during removal. The UI may be unstable.", plugin.get_display_name(),
                      exc_info=True)
        self.active_plugins.remove(plugin)

    #
    # Dispatchers
    #
    def _dispatch(self, func, sensitive, *args):
        for plugin in list(self.active_plugins):
            custom = getattr(plugin, func.__name__)
            if custom.__func__ is not func:
                try:
                    res = custom(*args)
                except Exception as e: #pylint: disable=broad-except
                    self._handle_error(plugin, func, sensitive, e)
                else:
                    yield res

    def _dispatch_with_plugin(self, func, sensitive, *args):
        for plugin in list(self.active_plugins):
            custom = getattr(plugin, func.__name__)
            if custom.__func__ is not func:
                try:
                    res = custom(*args)
                except Exception as e: #pylint: disable=broad-except
                    self._handle_error(plugin, func, sensitive, e)
                else:
                    yield plugin, res

    def _dispatch_single(self, plugin, func, sensitive, *args):
        custom = getattr(plugin, func.__name__)
        try:
            return custom(*args)
        except Exception as e:  #pylint: disable=broad-except
            self._handle_error(plugin, func, sensitive, e)
            return None

    def _handle_error(self, plugin, func, sensitive, exc):
        self.workspace.log("Plugin %s errored during %s" % (plugin.get_display_name(), func.__name__))
        self.workspace.log(exc)
        if sensitive:
            self.workspace.log("Deactivating %s for error during sensitive operation" % plugin.get_display_name())
            self.deactivate_plugin(plugin)

    def color_insn(self, addr, selected) -> Optional[QColor]:
        for res in self._dispatch(BasePlugin.color_insn, True, addr, selected):
            if res is not None:
                return res
        return None

    def color_block(self, addr) -> Optional[QColor]:
        for res in self._dispatch(BasePlugin.color_block, True, addr):
            if res is not None:
                return res
        return None

    def color_func(self, func) -> Optional[QColor]:
        for res in self._dispatch(BasePlugin.color_func, True, func):
            if res is not None:
                return res
        return None

    def draw_insn(self, qinsn, painter):
        for _ in self._dispatch(BasePlugin.draw_insn, True, qinsn, painter):
            pass

    def draw_block(self, qblock, painter):
        for _ in self._dispatch(BasePlugin.draw_block, True, qblock, painter):
            pass

    def instrument_disassembly_view(self, dview):
        for _ in self._dispatch(BasePlugin.instrument_disassembly_view, False, dview):
            pass

    def instrument_code_view(self, cview):
        for _ in self._dispatch(BasePlugin.instrument_code_view, False, cview):
            pass

    def handle_click_insn(self, qinsn, event):
        for res in self._dispatch(BasePlugin.handle_click_insn, False, qinsn, event):
            if res:
                return True
        return False

    def handle_click_block(self, qblock, event):
        for res in self._dispatch(BasePlugin.handle_click_block, False, qblock, event):
            if res:
                return True
        return False

    def handle_raise_view(self, view):
        for _ in self._dispatch(BasePlugin.handle_raise_view, False, view):
            pass

    def build_qblock_annotations(self, qblock: QBlock):
        for res in self._dispatch(BasePlugin.build_qblock_annotations, False, qblock):
            yield from res

    def build_context_menu_insn(self, insn):
        for res in self._dispatch(BasePlugin.build_context_menu_insn, False, insn):
            yield from res

    def build_context_menu_block(self, block):
        for res in self._dispatch(BasePlugin.build_context_menu_block, False, block):
            yield from res

    def build_context_menu_node(self, node):
        for res in self._dispatch(BasePlugin.build_context_menu_node, False, node):
            yield from res

    def build_context_menu_functions(self, funcs):
        for res in self._dispatch(BasePlugin.build_context_menu_functions, False, funcs):
            yield from res

    def get_func_column(self, idx):
        for plugin in self.active_plugins:
            if idx >= len(plugin.FUNC_COLUMNS):
                idx -= len(plugin.FUNC_COLUMNS)
            else:
                return plugin.FUNC_COLUMNS[idx]
        raise IndexError("Not enough columns")

    def count_func_columns(self):
        return sum((len(plugin.FUNC_COLUMNS) for plugin in self.active_plugins))

    def extract_func_column(self, func, idx):
        for plugin in self.active_plugins:
            if idx >= len(plugin.FUNC_COLUMNS):
                idx -= len(plugin.FUNC_COLUMNS)
            else:
                try:
                    return plugin.extract_func_column(func, idx)
                except Exception as e: #pylint: disable=broad-except
                    # this should really be a "sensitive" operation but like
                    self.workspace.log(e)
                    self.workspace.log("PLEASE FIX YOUR PLUGIN AHHHHHHHHHHHHHHHHH")
                    return 0, ''
        raise IndexError("Not enough columns")

    def step_callback(self, simgr):
        for _ in self._dispatch(BasePlugin.step_callback,True, simgr):
            pass

    def handle_stack_var_renamed(self, func, offset, old_name, new_name):
        for res in self._dispatch(BasePlugin.handle_stack_var_renamed, False, func, offset, old_name, new_name):
            if res:
                return True
        return False

    def handle_stack_var_retyped(self, func, offset, old_type, new_type):
        for res in self._dispatch(BasePlugin.handle_stack_var_retyped, False, func, offset, old_type, new_type):
            if res:
                return True
        return False

    def handle_func_arg_renamed(self, func, offset, old_name, new_name):
        for res in self._dispatch(BasePlugin.handle_func_arg_renamed, False, func, offset, old_name, new_name):
            if res:
                return True
        return False

    def handle_func_arg_retyped(self, func, offset, old_type, new_type):
        for res in self._dispatch(BasePlugin.handle_func_arg_retyped, False, func, offset, old_type, new_type):
            if res:
                return True
        return False

    def handle_global_var_renamed(self, address, old_name, new_name):
        for res in self._dispatch(BasePlugin.handle_global_var_renamed, False, address, old_name, new_name):
            if res:
                return True
        return False

    def handle_global_var_retyped(self, address, old_type, new_type):
        for res in self._dispatch(BasePlugin.handle_global_var_retyped, False, address, old_type, new_type):
            if res:
                return True
        return False

    def handle_other_var_renamed(self, var, old_name, new_name):
        for res in self._dispatch(BasePlugin.handle_other_var_renamed, False, var, old_name, new_name):
            if res:
                return True
        return False

    def handle_other_var_retyped(self, var, old_type, new_type):
        for res in self._dispatch(BasePlugin.handle_other_var_retyped, False, var, old_type, new_type):
            if res:
                return True
        return False

    def handle_function_renamed(self, func, old_name, new_name):
        for res in self._dispatch(BasePlugin.handle_function_renamed, False, func, old_name, new_name):
            if res:
                return True
        return False

    def handle_function_retyped(self, func, old_type, new_type):
        for res in self._dispatch(BasePlugin.handle_global_var_retyped, False, func, old_type, new_type):
            if res:
                return True
        return False

    def handle_comment_changed(self, address, old_cmt, new_cmt, created: bool, decomp: bool):
        for res in self._dispatch(BasePlugin.handle_comment_changed, False, address, old_cmt, new_cmt, created, decomp):
            if res:
                return True
        return False

    def handle_struct_changed(self, old_struct, new_struct):
        for res in self._dispatch(BasePlugin.handle_struct_changed, False, old_struct, new_struct):
            if res:
                return True
        return False

    def decompile_callback(self, func):
        for _ in self._dispatch(BasePlugin.decompile_callback, False, func):
            pass

    def handle_project_initialization(self):
        for _ in self._dispatch(BasePlugin.handle_project_initialization, False):
            pass

    def handle_project_save(self, file_name: str):
        for _ in self._dispatch(BasePlugin.handle_project_save, False, file_name):
            pass

    def on_workspace_initialized(self, workspace):
        for _ in self._dispatch(BasePlugin.on_workspace_initialized, False, workspace):
            pass

    def angrdb_store_entries(self):
        entries = {}
        for plugin, res in self._dispatch_with_plugin(BasePlugin.angrdb_store_entries, False):
            for result in res:
                if isinstance(result, tuple) and len(result) == 2:
                    key, value = result
                    key = plugin.__class__.__name__.split(".")[-1] + "___" + key
                    entries[key] = value
        return entries

    def angrdb_load_entries(self, entries):
        plugin_name_to_plugin = {}
        for plugin in list(self.active_plugins):
            plugin_name_to_plugin[plugin.__class__.__name__.split(".")[-1]] = plugin

        for key, value in entries.items():
            if "___" not in key:
                continue
            splitted = key.split("___")
            if len(splitted) != 2:
                continue
            plugin_class_name, key = splitted
            plugin = plugin_name_to_plugin.get(plugin_class_name, None)
            if plugin is None:
                continue
            # dispatch
            custom = getattr(plugin, "angrdb_load_entry")
            if custom is not BasePlugin.angrdb_load_entry:
                try:
                    custom(key, value)
                except Exception as ex:  # pylint: disable=broad-except
                    self._handle_error(plugin, BasePlugin.angrdb_load_entry, False, ex)



import os
import logging
from typing import Optional

from angr import SimState

try:
    import bintrace
    from bintrace.debugger_angr import AngrTraceDebugger
except ImportError as e:
    bintrace = None

from ...data.trace import BintraceTrace
from ...data.breakpoint import BreakpointType
from .debugger import Debugger


_l = logging.getLogger(name=__name__)


class BintraceDebugger(Debugger):
    """
    Trace playback debugger.
    """

    def __init__(self, trace: BintraceTrace, workspace: 'Workspace'):
        super().__init__(workspace)
        assert bintrace is not None
        assert isinstance(trace, BintraceTrace)
        self._trace: 'bintrace.Trace' = trace
        self._trace_dbg: AngrTraceDebugger = AngrTraceDebugger(self._trace.trace,
                                                               self.workspace.instance.project.am_obj)
        self._cached_simstate = None

    def __str__(self):
        pc = self.simstate.solver.eval(self.simstate.regs.pc)
        return f'{os.path.basename(self._trace.trace.path)} @ {pc:x}'

    def _on_state_change(self):
        """
        Common handler for state changes.
        """
        self._cached_simstate = None
        self.state_changed.am_event()

    def _sync_breakpoints(self):
        """
        Synchronize breakpoints set in Workspace with trace debugger.
        """
        bp_type_map = {
            BreakpointType.Execute: bintrace.debugger.BreakpointType.Execute,
            BreakpointType.Read: bintrace.debugger.BreakpointType.Read,
            BreakpointType.Write: bintrace.debugger.BreakpointType.Write,
        }
        self._trace_dbg.breakpoints = {
            bintrace.debugger.Breakpoint(bp_type_map[bp.type], bp.addr, bp.size)
            for bp in self.workspace.instance.breakpoint_mgr.breakpoints
        }

    @property
    def simstate(self) -> SimState:
        if self._cached_simstate is None:
            self._cached_simstate = self._trace_dbg.simstate
        return self._cached_simstate

    @property
    def is_running(self) -> bool:
        return True

    @property
    def can_step_backward(self) -> bool:
        return self._trace_dbg.can_step_backward

    def step_backward(self):
        if self.can_step_backward:
            self._trace_dbg.step_backward()
            self._on_state_change()

    @property
    def can_step_forward(self) -> bool:
        return self._trace_dbg.can_step_forward

    def step_forward(self, until_addr: Optional[int] = None):
        if self.can_step_forward:
            self._trace_dbg.step_forward(until_addr=until_addr)
            self._on_state_change()

    @property
    def can_continue_backward(self) -> bool:
        return self._trace_dbg.can_continue_backward

    def continue_backward(self):
        if self.can_continue_backward:
            self._sync_breakpoints()
            self._trace_dbg.continue_backward()
            self._on_state_change()

    @property
    def can_continue_forward(self) -> bool:
        return self._trace_dbg.can_continue_forward

    def continue_forward(self):
        if self.can_continue_forward:
            self._sync_breakpoints()
            self._trace_dbg.continue_forward()
            self._on_state_change()

    @property
    def can_halt(self) -> bool:
        return False  # XXX: Trace playback is "instantaneous", always is halted state.

    @property
    def is_halted(self) -> bool:
        return True

    @property
    def can_stop(self) -> bool:
        return True

    def stop(self):
        pass

    @property
    def is_exited(self) -> bool:
        return False

    def replay_to_nth_event(self, n: int):
        """
        Replay to the Nth event, skipping over stop events and ending on the nearest execution event.
        """
        t = self._trace.trace
        assert 0 <= n < t.get_num_events()

        until = t.get_nth_event(n)
        if until is None:
            _l.error('Could not seek to event %d', n)
            return

        until = t.get_prev_exec_event(until)
        if until is None:
            _l.error('No execution event prior to event %d', n)
            return

        self._trace_dbg.state = t.replay(self._trace_dbg.state, until)
        self._on_state_change()




import cle
import angr
import archinfo
try:
    import archr
except ImportError:
    archr = None

from .job import Job
from ...logic.threads import gui_thread_schedule
from ...ui.dialogs import LoadBinary


class LoadTargetJob(Job):
    """
    Job to load archr target and angr project.
    """

    def __init__(self, target, on_finish=None):
        super().__init__("Loading target", on_finish=on_finish)
        self.target = target

    def _run(self, inst):
        self._progress_callback(5)
        with self.target.build().start() as t:
            self._progress_callback(10)
            dsb = archr.arsenal.DataScoutBow(t)
            apb = archr.arsenal.angrProjectBow(t, dsb)
            partial_ld = apb.fire(return_loader=True, perform_relocations=False, load_debug_info=False)
            self._progress_callback(50)
            load_options, cfg_args, variable_recovery_args = gui_thread_schedule(LoadBinary.run, (partial_ld,))
            if cfg_args is None:
                return

            # Create the project, load it, then record the image name on success
            proj = apb.fire(use_sim_procedures=True, load_options=load_options)
            self._progress_callback(95)
            inst._reset_containers()
            inst.project = proj
            inst.project.am_event(cfg_args=cfg_args, variable_recovery_args=variable_recovery_args)


class LoadBinaryJob(Job):
    """
    Job to display binary load dialog and create angr project.
    """

    def __init__(self, fname, load_options=None, on_finish=None):
        super().__init__("Loading file", on_finish=on_finish)
        self.load_options = load_options or {}
        self.fname = fname

    def _run(self, inst):
        self._progress_callback(5)

        partial_ld = None
        try:
            # Try automatic loading
            partial_ld = cle.Loader(self.fname, perform_relocations=False, load_debug_info=False)
        except archinfo.arch.ArchNotFound as e:
            partial_ld = cle.Loader(self.fname, perform_relocations=False, load_debug_info=False, arch='x86')
            gui_thread_schedule(LoadBinary.binary_arch_detect_failed, (self.fname, str(e)))
        except cle.CLECompatibilityError:
            # Continue loading as blob
            pass

        if partial_ld is None:
            try:
                # Try loading as blob; dummy architecture (x86) required, user will select proper arch
                partial_ld = cle.Loader(self.fname, main_opts={'backend': 'blob', 'arch': 'x86'})
            except cle.CLECompatibilityError:
                # Failed to load executable, even as blob!
                gui_thread_schedule(LoadBinary.binary_loading_failed, (self.fname,))
                return

        self._progress_callback(50)
        new_load_options, cfg_args, variable_recovery_args = gui_thread_schedule(LoadBinary.run, (partial_ld, ))
        if cfg_args is None:
            return

        engine = None
        if hasattr(new_load_options['arch'], 'pcode_arch'):
            engine = angr.engines.UberEnginePcode

        self.load_options.update(new_load_options)

        proj = angr.Project(self.fname, load_options=self.load_options, engine=engine)
        self._progress_callback(95)
        def callback():
            inst._reset_containers()
            inst.project.am_obj = proj
            inst.project.am_event(cfg_args=cfg_args, variable_recovery_args=variable_recovery_args)
        gui_thread_schedule(callback, ())
